/**********************************************/
/* CPU_DataMap.cpp                            */
/*--------------------------------------------*/
/* POTATO Chip Data Memory Map Handlers       */
/*--------------------------------------------*/
/* Copyright (C) 2002 Michel Gerritse         */
/*                                            */
/* Last update: 03-10-2002                    */
/*--------------------------------------------*/
/* TODO:                                      */
/**********************************************/
#include "VMUCore.h"
#include "DreamVMU.h"

unsigned char BiosRom[0x10000];
unsigned char FlashRom[2][0x10000];
unsigned char Ram[2][0x100];
unsigned char WorkRam[2][0x100];
unsigned char XRam[3][0x80];

unsigned char SFR[0x100];

unsigned char ReadByte(unsigned short Addr)
{
  if (Addr < 0x100) /* Ram */
  {
    return Ram[Cpu.RAMBK0][Addr];
  }
  else if (Addr >= 0x180) /* XRAM */
  {
    return XRam[Cpu.XBNK][Addr - 0x180];
  }
  else /* SFR */
  {
    switch (Addr)
    {
      case 0x100: /* Accumulator */
        return Cpu.Acc;

      case 0x101: /* Program Status Word */
        Cpu.P = Parity[Cpu.Acc];
        GetPSW();
        
        return Cpu.PSW;

      case 0x102: /* B Register */
        return Cpu.B;

      case 0x103: /* C Register */
        return Cpu.C;

      case 0x104: /* Table Reference Register Low */
        return Cpu.TRR.Byte.Low;

      case 0x105: /* Table Reference Register High */
        return Cpu.TRR.Byte.High;

      case 0x106: /* Stack Pointer */
        return Cpu.SP;

      case 0x107: /* Power Control Regitser */
        return Cpu.PCON;

      case 0x108: /* Master Interrupt Enable Control Register */
        return Cpu.IE;

      case 0x109: /* Interrupt Priority Control Register */
        return Cpu.IP;

      case 0x10D: /* External Memory Control Register */
        return Cpu.EXT;

      case 0x10E: /* Oscillation Control Regitser */
        return Cpu.OCR;

      case 0x110: /* Timer 0 Control Register */
        return Cpu.Timer0.T0CNT;

      case 0x111: /* Timer 0 Prescale Data Register */
        return Cpu.Timer0.T0PRR;

      case 0x112: /* Timer 0 Low Register */
        return Cpu.Timer0.Timer.Byte.T0L;

      case 0x113: /* Timer 0 Low Reload Register */
        return Cpu.Timer0.Reload.Byte.T0LR;

      case 0x114: /* Timer 0 High Register */
        return Cpu.Timer0.Timer.Byte.T0H;

      case 0x115: /* Timer 0 High Reload Register */
        return Cpu.Timer0.Reload.Byte.T0HR;

      case 0x120: /* Mode Control Register */
        return Cpu.MCR;

      case 0x125: /* Bank Address Register */
        return Cpu.XBNK;

      case 0x127: /* LCD Contrast Control Register */
        return Cpu.VCCR;

      case 0x144: /* Port 1 Latch */
        return Cpu.P1;

      case 0x14C: /* Port 3 Latch */
        return Cpu.P3;

      case 0x14E: /* Port 3 Interrupt Control Register */
        return Cpu.P3INT;

      case 0x154: /* FlashRom Programming Register */
        return Cpu.FPR;

      case 0x15C: /* Port 7 Latch */
        return Cpu.P7;

      case 0x163: /* VMU Control Register */
        return Cpu.VSEL;

      case 0x164: /* Work Ram Access Address 1 */
        return Cpu.VRMAD1;

      case 0x165: /* Work Ram Access Address 2 */
        return Cpu.VRMAD2;

      case 0x166: /* Send/Recieve Buffer */
      {
        unsigned char Data = WorkRam[Cpu.VRMAD2][Cpu.VRMAD1];

        if (Cpu.VSEL & 0x10) /* Auto Increment */
        {
          if (++Cpu.VRMAD1 == 0x00)
          {
            /* Switch Work Ram Bank */
            Cpu.VRMAD2 ^= 1;
          }
        }
        
        return Data;
      }

      case 0x17F: /* Base Timer Control Register */
        return Cpu.BaseTimer.BTCR;
    }
  }

  //ShowMessage("SFR 0x%x Read", Addr);
     
  return SFR[Addr & 0xFF];
}

void WriteByte(unsigned short Addr, unsigned char Data)
{
  if (Addr < 0x100) /* Ram */
  {
    Ram[Cpu.RAMBK0][Addr] = Data;
    return;
  }
  else if (Addr >= 0x180) /* XRAM */
  {
    XRam[Cpu.XBNK][Addr - 0x180] = Data;
    return;
  }
  else /* SFR */
  {
    switch (Addr)
    {
      case 0x100: /* Accumulator */
        Cpu.Acc = Data;
        return;

      case 0x101: /* Program Status Word */
        Cpu.PSW = Data;
        SetPSW();
        return;

      case 0x102: /* B Register */
        Cpu.B = Data;
        return;

      case 0x103: /* C Register */
        Cpu.C = Data;
        return;

      case 0x104: /* Table Reference Register Low */
        Cpu.TRR.Byte.Low = Data;
        return;

      case 0x105: /* Table Reference Register High */
        Cpu.TRR.Byte.High = Data;
        return;

      case 0x106: /* Stack Pointer */
        Cpu.SP = Data;
        return;

      case 0x107: /* Power Control Regitser */
        Cpu.PCON = Data;

        Cpu.Hold = Data & 0x2;
        Cpu.Halt = Data & 0x1;
        return;

      case 0x108: /* Master Interrupt Enable Control Register */
        Cpu.IE = Data;
        return;

      case 0x109: /* Interrupt Priority Control Register */
        Cpu.IP = Data;
        return;

      case 0x10D: /* External Memory Control Register */
        Cpu.EXT = Data;
        return;

      case 0x10E: /* Oscillation Control Regitser */
      {
        Cpu.OCR = Data;

        int OldSpeed = Cpu.Speed;

        switch (Data & 0xA0)
        {
          case 0x00:
            Cpu.Speed = 5000; /* (600Khz / 2) / 60 (RC Clock / 12) */
            break;
          case 0x20:     
            Cpu.Speed = 273; /* (32768Hz / 2) / 60 (Sub Clock / 12) */
            break;
          case 0x80:
            Cpu.Speed = 10000; /* (600Khz / 1) / 60 (RC Clock / 6) */
            break;
          case 0xA0:
            Cpu.Speed = 546; /* (32768Hz / 1) / 60 (Sub Clock / 6-) */
            break;
        }

        /* Calculate remaining cycles to run, based on new clock speed */
        Cpu.NrCycles = Cpu.Speed * (Cpu.NrCycles / OldSpeed);
        return;
      }

      case 0x110: /* Timer 0 Control Register */
        Cpu.Timer0.T0CNT = Data;

        /* Reload High Data */
        if ((Data & 0x80) == 0)
        {
          Cpu.Timer0.Timer.Byte.T0H = Cpu.Timer0.Reload.Byte.T0HR;
        }
        
        /* Reload Low Data */
        if ((Data & 0x40) == 0)
        {
          Cpu.Timer0.Timer.Byte.T0L = Cpu.Timer0.Reload.Byte.T0LR;
        }
        return;

      case 0x111: /* Timer 0 Prescale Data Register */
        Cpu.Timer0.T0PRR = Data;
        Cpu.Timer0.PreScale = (256 - Data);
        return;

      case 0x113: /* Timer 0 Low Reload Register */
        Cpu.Timer0.Reload.Byte.T0LR = Data;
        
        if ((Cpu.Timer0.T0CNT & 0x40) == 0)
        {
          Cpu.Timer0.Timer.Byte.T0L = Data;
        }

        return;

      case 0x115: /* Timer 0 High Reload Register */
        Cpu.Timer0.Reload.Byte.T0HR = Data;
        
        if ((Cpu.Timer0.T0CNT & 0x80) == 0)
        {
          Cpu.Timer0.Timer.Byte.T0H = Data;
        }

        return;

      case 0x120: /* Mode Control Register */
        Cpu.MCR = Data;
        return;

      case 0x125: /* Bank Address Register */
        Cpu.XBNK = Data;
        return;

      case 0x127: /* LCD Contrast Control Register */
        Cpu.VCCR = Data;
        return;

      case 0x144: /* Port 1 Latch */
        Cpu.P1 = Data;
        return;

      case 0x14C: /* Port 3 Latch */
        Cpu.P3 = Data;
        return;

      case 0x14E: /* Port 3 Interrupt Control Register */
        Cpu.P3INT = Data;
        return;

      case 0x154: /* FlashRom Programming Register */
        Cpu.FPR = Data;
        return;

      case 0x163: /* VMU Control Register */
        Cpu.VSEL = Data;
        return;

      case 0x164: /* Work Ram Access Address 1 */
        Cpu.VRMAD1 = Data;
        return;

      case 0x165: /* Work Ram Access Address 2 */
        Cpu.VRMAD2 = Data & 1;
        return;

      case 0x166: /* Send/Recieve Buffer */
        WorkRam[Cpu.VRMAD2][Cpu.VRMAD1] = Data;

        if (Cpu.VSEL & 0x10) /* Auto Increment */
        {
          if (++Cpu.VRMAD1 == 0x00)
          {
            /* Switch Work Ram Bank */
            Cpu.VRMAD2 ^= 1;
          }
        }
        return;

      case 0x17F: /* Base Timer Control Register */
        Cpu.BaseTimer.BTCR = Data;

        /* Timer stopped */
        if ((Data & 0x40) == 0)
        {
          /* Reset Counter */
          Cpu.BaseTimer.Counter = 0x0000;
        }
        return;

    }
  }

  //ShowMessage("SFR 0x%x Write 0x%x", Addr, Data);

  SFR[Addr & 0xFF] = Data;
}