/**********************************************/
/* CPU_Opcodes.cpp                            */
/*--------------------------------------------*/
/* POTATO Chip instruction set implementation */
/*--------------------------------------------*/
/* Copyright (C) 2002 Michel Gerritse         */
/*                                            */
/* Last update: 16-10-2002                    */
/*--------------------------------------------*/
/* TODO:                                      */
/*      - Port latch Read/Writes              */
/*      - Fix OV flag calculation (ADD/SUB)   */
/**********************************************/
#include "VMUCore.h"
#include "DreamVMU.h"

/* Flag calculations */
#define SetOV_Add(x, y, z) (Cpu.OV = (((x) ^ (y)) & ((x) ^ (z)) & 0x80) ? 1 : 0)
#define SetOV_Sub(x, y, z) (Cpu.OV = (((z) ^ (y)) & ((z) ^ (x)) & 0x80) ? 1 : 0)

#define SetCY_Add(y, z)    (Cpu.CY = (((y) + (z)) > 0xFF) ? 1 : 0)
#define SetCY_Sub(y, z)    (Cpu.CY = (((y) - (z)) < 0x00) ? 1 : 0)

#define SetAC_Add(y, z)    (Cpu.AC = ((((y) & 0xF) + ((z) & 0xF)) > 0xF) ? 1 : 0)
#define SetAC_Sub(y, z)    (Cpu.AC = ((((y) & 0xF) - ((z) & 0xF)) < 0x0) ? 1 : 0)

unsigned short GetIndirectAddr(unsigned char Rj)
{
  unsigned short Addr;

  if (Rj & 0x2) /* SFR Access */
  {
    Addr = 0x100;
  }
  else /* RAM Access */
  {
    Addr = 0x00;
  }

  Addr += Ram[Cpu.RAMBK0][(Cpu.IRBK1 << 3) | (Cpu.IRBK0 << 2) | Rj];

  return Addr;
}

/***********************************************/
/****** Arithmetic Operation Instructions ******/
/***********************************************/

/* ADD #i8 (ADD immediate data to accumulator) */
/***********************************************/
/* OpCode      : 0x81h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- (ACC) + #i8          */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1];
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* ADD d9 (ADD direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0x82 - 0x83h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- (ACC) + @d9          */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8 ) | Cpu.Code[Cpu.PC + 1]);
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* ADD @Rj (ADD indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0x84 - 0x87h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)=(ACC)+((Rj)) j=0,1,2,3  */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDrj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3));
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);

  Cpu.PC += 1;
  return 1;
}

/* ADDC #i8 (ADD indirect data and carry flag to accumulator)  */
/***********************************************/
/* OpCode      : 0x91h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) + (CY) + #i8    */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDCi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1] + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);
  
  Cpu.PC += 2;
  return 1;
}

/* ADDC d9 (ADD direct byte and carry flag to accumulator) */
/***********************************************/
/* OpCode      : 0x92 - 0x93h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) + (CY) + @d9    */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDCd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8 ) | Cpu.Code[Cpu.PC + 1]) + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* ADDC @Rj (ADD indirect byte and carry flag to accumulator) */
/***********************************************/
/* OpCode      : 0x94 - 0x97h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)=(ACC)+(CY)+((Rj))       */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int ADDCrj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3)) + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc += Data;
  
  SetOV_Add(Cpu.Acc, OldAcc, Data);
  SetCY_Add(OldAcc, Data);
  SetAC_Add(OldAcc, Data);

  Cpu.PC += 1;
  return 1;
}

/* SUB #i8 (Subtract immediate data from accumulator) */
/***********************************************/
/* OpCode      : 0xA1h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- (ACC) - #i8          */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1];
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* SUB d9 (Suntract direct byte from accumulator) */
/***********************************************/
/* OpCode      : 0xA2 - 0xA3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- (ACC) - @d9          */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8 ) | Cpu.Code[Cpu.PC + 1]);
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* SUB @Rj (Subtract indirect byte from accumulator) */
/***********************************************/
/* OpCode      : 0xA4 - 0xA7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)=(ACC)-((Rj)) j=0,1,2,3  */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBrj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3));
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 1;
  return 1;
}

/* SUBC #i8 (Subtract indirect data and carry flag from accumulator)  */
/***********************************************/
/* OpCode      : 0xB1h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) - ((CY) + #i8)  */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBCi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1] + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* SUBC d9 (Subtract direct byte and carry flag to accumulator) */
/***********************************************/
/* OpCode      : 0xB2 - 0xB3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) - ((CY) + @d9)  */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBCd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8 ) | Cpu.Code[Cpu.PC + 1]) + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 2;
  return 1;
}

/* SUBC @Rj (Subtract indirect byte and carry flag to accumulator) */
/***********************************************/
/* OpCode      : 0xB4 - 0xB7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)=(ACC)- ((CY)+((Rj)))    */
/* Flags Used  : CY*AC*OV                      */
/***********************************************/
int SUBCrj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3)) + Cpu.CY;
  unsigned char OldAcc = Cpu.Acc;
  
  Cpu.Acc -= Data;
  
  SetOV_Sub(Cpu.Acc, OldAcc, Data);
  SetCY_Sub(OldAcc, Data);
  SetAC_Sub(OldAcc, Data);

  Cpu.PC += 1;
  return 1;
}

/* INC d9 (Increment direct byte) */
/***********************************************/
/* OpCode      : 0x62 - 0x63h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9) = (d9) + 1               */
/* Flags Used  : None                          */
/***********************************************/
int INCd9(unsigned char Opcode)
{
  unsigned short Addr = ((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1];

  WriteByte(Addr, ReadByte(Addr) + 1);

  Cpu.PC += 2;
  return 1;
}

/* INC @Rj (Increment indirect byte) */
/***********************************************/
/* OpCode      : 0x64 - 0x67h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : ((Rj) = ((Rj)) + 1            */
/* Flags Used  : None                          */
/***********************************************/
int INCrj(unsigned char Opcode)
{
  unsigned short Addr = GetIndirectAddr(Opcode & 0x3);
  WriteByte(Addr, ReadByte(Addr) + 1);
  
  Cpu.PC += 1;
  return 1;
}

/* DEC d9 (Decrement direct byte) */
/***********************************************/
/* OpCode      : 0x72 - 0x73h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9) = (d9) - 1               */
/* Flags Used  : None                          */
/***********************************************/
int DECd9(unsigned char Opcode)
{
  unsigned short Addr = ((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1];

  WriteByte(Addr, ReadByte(Addr) - 1);

  Cpu.PC += 2;
  return 1;
}

/* DEC @Rj (Decrement indirect byte) */
/***********************************************/
/* OpCode      : 0x74 - 0x77h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : ((Rj) = ((Rj)) - 1            */
/* Flags Used  : None                          */
/***********************************************/
int DECrj(unsigned char Opcode)
{
  unsigned short Addr = GetIndirectAddr(Opcode & 0x3);
  WriteByte(Addr, ReadByte(Addr) - 1);
  
  Cpu.PC += 1;
  return 1;
}

/* MUL (Multiply accumulator and c register times b register) */
/***********************************************/
/* OpCode      : 0x30h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 7                             */
/* Function    : (B)(ACC)(C) = (ACC)(C) * (B)  */
/* Flags Used  : CY*OV                         */
/***********************************************/
int MUL(unsigned char Opcode)
{
  unsigned int Result = ((Cpu.Acc << 8) | Cpu.C) * Cpu.B;
  Cpu.B   = (Result & 0xFF0000) >> 16;
  Cpu.Acc = (Result & 0x00FF00) >> 8;
  Cpu.C   = (Result & 0x0000FF);

  /* Set OV if B != 0, else reset OV */
  SetOV((Cpu.B == 0) ? 0 : 1);
  
  /* CY is always reset */
  SetCY(0);

  Cpu.PC += 1;
  return 7;
}

/* DIV (Divide accumulator and c register by b register) */
/***********************************************/
/* OpCode      : 0x40h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 7                             */
/* Function    : (ACC)(C)mod(B)=(ACC)(C) / (B) */
/* Flags Used  : CY*OV                         */
/***********************************************/
int DIV(unsigned char Opcode)
{
  if (Cpu.B == 0) /* Divide by Zero */
  {
    Cpu.Acc = 0xFF;

    SetOV(1);
    SetCY(0);
  }
  else
  {
    unsigned short Data     = ((Cpu.Acc << 8) | Cpu.C);
    unsigned short Quotient = Data / Cpu.B;
    
    Cpu.Acc = (Quotient >> 8);
    Cpu.C   = (Quotient & 0xFF);
    Cpu.B   = (Data % Cpu.B);

    /* Set OV if B == 0, else reset OV */
    SetOV((Cpu.B == 0) ? 1 : 0);
  
    /* CY is always reset */
    SetCY(0);
  }

  Cpu.PC += 1;
  return 7;
}

/********************************************/
/****** Logical Operation Instructions ******/
/********************************************/

/* AND #i8 (AND immediate data to accumulator) */
/***********************************************/
/* OpCode      : 0xE1h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) & #i8           */
/* Flags Used  : None                          */
/***********************************************/
int ANDi8(unsigned char Opcode)
{
  Cpu.Acc &= Cpu.Code[Cpu.PC + 1];

  Cpu.PC += 2;
  return 1;
}

/* AND d9 (AND direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0xE2 - 0xE3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) & (d9)          */
/* Flags Used  : None                          */
/***********************************************/
int ANDd9(unsigned char Opcode)
{
  Cpu.Acc &= ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 1;
}

/* AND @Rj (AND indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0xE4 - 0xE7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) & ((Rj))        */
/* Flags Used  : None                          */
/***********************************************/
int ANDrj(unsigned char Opcode)
{
  Cpu.Acc &= ReadByte(GetIndirectAddr(Opcode & 0x3));

  Cpu.PC += 1;
  return 1;
}

/* OR #i8 (OR immediate data to accumulator) */
/***********************************************/
/* OpCode      : 0xD1h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) | #i8           */
/* Flags Used  : None                          */
/***********************************************/
int ORi8(unsigned char Opcode)
{
  Cpu.Acc |= Cpu.Code[Cpu.PC + 1];

  Cpu.PC += 2;
  return 1;
}

/* OR d9 (OR direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0xD2 - 0xD3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) | (d9)          */
/* Flags Used  : None                          */
/***********************************************/
int ORd9(unsigned char Opcode)
{
  Cpu.Acc |= ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 1;
}

/* OR @Rj (OR indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0xD4 - 0xD7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) | ((Rj))        */
/* Flags Used  : None                          */
/***********************************************/
int ORrj(unsigned char Opcode)
{
  Cpu.Acc |= ReadByte(GetIndirectAddr(Opcode & 0x3));

  Cpu.PC += 1;
  return 1;
}

/* XOR #i8 (XOR immediate data to accumulator) */
/***********************************************/
/* OpCode      : 0xF1h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) ^ #i8           */
/* Flags Used  : None                          */
/***********************************************/
int XORi8(unsigned char Opcode)
{
  Cpu.Acc ^= Cpu.Code[Cpu.PC + 1];

  Cpu.PC += 2;
  return 1;
}

/* XOR d9 (XOR direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0xF2 - 0xF3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) ^ (d9)          */
/* Flags Used  : None                          */
/***********************************************/
int XORd9(unsigned char Opcode)
{
  Cpu.Acc ^= ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 1;
}

/* XOR @Rj (XOR indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0xF4 - 0xF7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) = (ACC) ^ ((Rj))        */
/* Flags Used  : None                          */
/***********************************************/
int XORrj(unsigned char Opcode)
{
  Cpu.Acc ^= ReadByte(GetIndirectAddr(Opcode & 0x3));

  Cpu.PC += 1;
  return 1;
}

/* ROL (Rotate accumulator left) */
/***********************************************/
/* OpCode      : 0xE0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : A7<-A6<-A5<-A4<-A3<-A2<-A1<-A0*/
/* Flags Used  : None                          */
/***********************************************/
int ROL(unsigned char Opcode)
{
  if (Cpu.Acc & 0x80)
  {
    Cpu.Acc <<= 1;
    Cpu.Acc |= 0x1;
  }
  else
  {
    Cpu.Acc <<= 1;
  }

  Cpu.PC += 1;
  return 1;
}

/* ROLC (Rotate accumulator left through the carry flag) */
/***********************************************/
/* OpCode      : 0xF0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : A7<-A6<-A5...A2<-A1<-A0<-CY<- */
/* Flags Used  : CY                            */
/***********************************************/
int ROLC(unsigned char Opcode)
{
  if (Cpu.Acc & 0x80)
  {
    Cpu.Acc <<= 1;
    Cpu.Acc |= Cpu.CY;
    SetCY(1);
  }
  else
  {
    Cpu.Acc <<= 1;
    Cpu.Acc |= Cpu.CY;
    SetCY(0);
  }

  Cpu.PC += 1;
  return 1;
}

/* ROR (Rotate accumulator right) */
/***********************************************/
/* OpCode      : 0xC0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : A7->A6->A5->A4->A3->A2->A1->A0*/
/* Flags Used  : None                          */
/***********************************************/
int ROR(unsigned char Opcode)
{
  if (Cpu.Acc & 0x1)
  {
    Cpu.Acc >>= 1;
    Cpu.Acc |= 0x80;
  }
  else
  {
    Cpu.Acc >>= 1;
  }

  Cpu.PC += 1;
  return 1;
}

/* RORC (Rotate accumulator right through the carry flag) */
/***********************************************/
/* OpCode      : 0xD0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : ->CY->A7->A6->A5...A2->A1->A0 */
/* Flags Used  : CY                            */
/***********************************************/
int RORC(unsigned char Opcode)
{
  if (Cpu.Acc & 0x1)
  {
    Cpu.Acc >>= 1;
    Cpu.Acc |= (Cpu.CY << 7);
    SetCY(1);
  }
  else
  {
    Cpu.Acc >>= 1;
    Cpu.Acc |= (Cpu.CY << 7);
    SetCY(0);
  }

  Cpu.PC += 1;
  return 1;
}

/****************************************/
/****** Data Transfer Instructions ******/
/****************************************/

/* LD d9 (Load direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0x02 - 0x03h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- (d9)                 */
/* Flags Used  : None                          */
/***********************************************/
int LDd9(unsigned char Opcode)
{
  Cpu.Acc = ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 1;
}

/* LD @Rj (Load indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0x04 - 0x07h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC) <- ((Rj))               */
/* Flags Used  : None                          */
/***********************************************/
int LDrj(unsigned char Opcode)
{
  Cpu.Acc = ReadByte(GetIndirectAddr(Opcode & 3));

  Cpu.PC += 1;
  return 1;
}

/*ST d9 (Store direct byte to accumulator) */
/***********************************************/
/* OpCode      : 0x12 - 0x13h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9) <- (ACC)                 */
/* Flags Used  : None                          */
/***********************************************/
int STd9(unsigned char Opcode)
{
  WriteByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1], Cpu.Acc);

  Cpu.PC += 2;
  return 1;
}

/* ST @Rj (Store indirect byte to accumulator) */
/***********************************************/
/* OpCode      : 0x14 - 0x17h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : ((Rj)) <- (ACC)               */
/* Flags Used  : None                          */
/***********************************************/
int STrj(unsigned char Opcode)
{
  WriteByte(GetIndirectAddr(Opcode & 3), Cpu.Acc);

  Cpu.PC += 1;
  return 1;
}

/* MOV #i8,d9 (Move immediate data to direct byte) */
/***********************************************/
/* OpCode      : 0x22 - 0x23h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (d9) <- #i8                   */
/* Flags Used  : None                          */
/***********************************************/
int MOVi8d9(unsigned char Opcode)
{
  WriteByte((((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]), Cpu.Code[Cpu.PC + 2]);

  Cpu.PC += 3;
  return 2;
}

/* MOV #i8,@Rj (Move immediate data to indirect byte) */
/***********************************************/
/* OpCode      : 0x24 - 0x27h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : ((Rj)) <- #i8                 */
/* Flags Used  : None                          */
/***********************************************/
int MOVi8rj(unsigned char Opcode)
{
  WriteByte(GetIndirectAddr(Opcode & 0x3), Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 1;
}

/* LDC (Load code byte relative to TRR to accumulator) */
/***********************************************/
/* OpCode      : 0xC1h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 2                             */
/* Function    : (ACC)<-(BNK)(TRR)+(ACC) (ROM) */
/* Flags Used  : None                          */
/***********************************************/
int LDC(unsigned char Opcode)
{
  unsigned short Addr = Cpu.TRR.Value + Cpu.Acc;
  
  if (Cpu.EXT & 1)
    Cpu.Acc = FlashRom[0][Addr];
  else
    Cpu.Acc = BiosRom[Addr];

  Cpu.PC += 1;
  return 2;
}

/* LDF (Load code byte relative to TRR to accumulator) */
/***********************************************/
/* OpCode      : 0x50h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 2                             */
/* Function    : (ACC)<-(BNK)(TRR)   (ROM)     */
/* Flags Used  : None                          */
/***********************************************/
int LDF(unsigned char Opcode)
{
  Cpu.Acc = FlashRom[Cpu.FPR & 1][Cpu.TRR.Value];

  Cpu.PC += 1;
  return 2;
}

/* STF (Store code byte relative to TRR from accumulator) */
/***********************************************/
/* OpCode      : 0x51h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 2                             */
/* Function    : (BNK)(TRR)<-(ACC)   (ROM)     */
/* Flags Used  : None                          */
/***********************************************/
int STF(unsigned char Opcode)
{
  if ((Cpu.FPR & 2) == 0)
  {
    FlashRom[Cpu.FPR & 1][Cpu.TRR.Value] = Cpu.Acc;
  }

  Cpu.PC += 1;
  return 2;
}

/* PUSH d9 (Push direct byte to stack) */
/***********************************************/
/* OpCode      : 0x60 - 0x61h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (SP)<-(SP)+1,((SP))<-(d9)     */
/* Flags Used  : None                          */
/***********************************************/
int PUSH(unsigned char Opcode)
{
  Ram[0][++Cpu.SP] = ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);

  Cpu.PC += 2;
  return 2;
}

/* POP d9 (Pop direct byte from stack) */
/***********************************************/
/* OpCode      : 0x70 - 0x71h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (d9)<-((SP)),(SP)<-(SP)-1     */
/* Flags Used  : None                          */
/***********************************************/
int POP(unsigned char Opcode)
{
  WriteByte((((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]), Ram[0][Cpu.SP--]);

  Cpu.PC += 2;
  return 2;
}

/* XCH d9 (Exchange direct byte with accumulator) */
/***********************************************/
/* OpCode      : 0xC2 - 0xC3h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)<->(d9)                  */
/* Flags Used  : None                          */
/***********************************************/
int XCHd9(unsigned char Opcode)
{
  unsigned short Addr = (((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);
  unsigned char Data = ReadByte(Addr);
  WriteByte(Addr, Cpu.Acc);
  Cpu.Acc = Data;

  Cpu.PC += 2;
  return 1;
}

/* XCH @Rj (Exchange indirect byte with accumulator) */
/***********************************************/
/* OpCode      : 0xC4 - 0xC7h                  */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    : (ACC)<->(d9)                  */
/* Flags Used  : None                          */
/***********************************************/
int XCHrj(unsigned char Opcode)
{
  unsigned short Addr = GetIndirectAddr(Opcode & 0x3);
  unsigned char Data = ReadByte(Addr);
  WriteByte(Addr, Cpu.Acc);
  Cpu.Acc = Data;

  Cpu.PC += 1;
  return 1;
}

/*******************************/
/****** Jump Instructions ******/
/*******************************/

/* JMP a12 (Jump near absolute address) */
/***********************************************/
/* OpCode      : 0x28 - 0x2Fh, 0x38 - 0x3Fh    */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+2, (PC11-00)<-a12  */
/* Flags Used  : None                          */
/***********************************************/
int JMP(unsigned char Opcode)
{
  unsigned short Data = (((Opcode & 0x10) >> 1) | Opcode & 0x7) << 8;
  Data |= Cpu.Code[Cpu.PC + 1];
  
  Cpu.PC += 2;
  Cpu.PC = (Cpu.PC & 0xF000) | (Data & 0x0FFF);

  return 2;
}

/* JMPF a16 (Jump far absolute address) */
/***********************************************/
/* OpCode      : 0x21h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-a16                     */
/* Flags Used  : None                          */
/***********************************************/
int JMPF(unsigned char Opcode)
{
  Cpu.PC = (Cpu.Code[Cpu.PC + 1] << 8) | Cpu.Code[Cpu.PC + 2];

  if (Cpu.EXT & 1)
  {
    Cpu.Code = FlashRom[0];
  }
  else
  {
    Cpu.Code = BiosRom;
  }

  return 2;
}

/* BR r8 (Branch near relative address) */
/***********************************************/
/* OpCode      : 0x01h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+2,(PC)<-(PC)+r8    */
/* Flags Used  : None                          */
/***********************************************/
int BR(unsigned char Opcode)
{
  Cpu.PC = Cpu.PC + 2 + (char)(Cpu.Code[Cpu.PC + 1]);
  
  return 2;
}

/* BRF r16 (Branch far relative address) */
/***********************************************/
/* OpCode      : 0x11h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 4                             */
/* Function    : (PC)<-(PC)+3,(PC)<-(PC)-1+r16 */
/* Flags Used  : None                          */
/***********************************************/
int BRF(unsigned char Opcode)
{
  Cpu.PC = Cpu.PC + 2 + ((Cpu.Code[Cpu.PC + 2] << 8) | Cpu.Code[Cpu.PC + 1]);
  
  return 4;
}

/************************************************/
/****** Conditional Branching Instructions ******/
/************************************************/

/* BZ r8 (Branch near relative address if accumulator is zero) */
/***********************************************/
/* OpCode      : 0x80h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+2,if(ACC)=0        */
/*             : then (PC) + r8                */
/* Flags Used  : None                          */
/***********************************************/
int BZ(unsigned char Opcode)
{
  if (Cpu.Acc == 0)
  {
    Cpu.PC = Cpu.PC + 2 + (char)Cpu.Code[Cpu.PC + 1];
  }
  else
  {
    Cpu.PC += 2;
  }

  return 2;
}

/* BNZ r8 (Branch near relative address if accumulator is not zero) */
/***********************************************/
/* OpCode      : 0x90h                         */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+2,if(ACC)!=0       */
/*             : then (PC) + r8                */
/* Flags Used  : None                          */
/***********************************************/
int BNZ(unsigned char Opcode)
{
  if (Cpu.Acc != 0)
  {
    Cpu.PC = Cpu.PC + 2 + (char)Cpu.Code[Cpu.PC + 1];
  }
  else
  {
    Cpu.PC += 2;
  }

  return 2;
}

/* BP d9,b3,r8 (Branch near relative address if direct bit is possitive) */
/***********************************************/
/* OpCode      : 0x68 - 0x6Fh, 0x78 - 0x7Fh    */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(d9,b3)=1      */
/*             : then (PC)<-(PC) + r8          */
/* Flags Used  : None                          */
/***********************************************/
int BP(unsigned char Opcode)
{
  unsigned short OldPC = Cpu.PC + 1;
  int Bit = Opcode & 0x7;
  unsigned char Data = ReadByte(((Opcode & 0x10) << 4) | Cpu.Code[OldPC]);

  if (Data & (1 << Bit))
  {
    Cpu.PC += 3;
    Cpu.PC += (char)Cpu.Code[OldPC + 1];
  }
  else
  {
    Cpu.PC += 3;
  }

  return 2;
}

/* BPC d9,b3,r8 (Branch near relative address if direct bit is possitive, and clear) */
/***********************************************/
/* OpCode      : 0x48 - 0x4Fh, 0x58 - 0x5Fh    */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(d9,b3)=1      */
/*             : then (PC)<-(PC)+r8,(d9,b3)=0  */
/* Flags Used  : None                          */
/***********************************************/
int BPC(unsigned char Opcode)
{
  unsigned short OldPC = Cpu.PC + 1;
  int Bit = Opcode & 0x7;
  unsigned char Data = ReadByte(((Opcode & 0x10) << 4) | Cpu.Code[OldPC]);

  if (Data & (1 << Bit))
  {
    Cpu.PC += 3;
    Cpu.PC += (char)Cpu.Code[OldPC + 1];

    WriteByte(Cpu.Code[OldPC], Data ^ (1 << Bit));
  }
  else
  {
    Cpu.PC += 3;
  }

  return 2;
}

/* BN d9,b3,r8 (Branch near relative address if direct bit is negative) */
/***********************************************/
/* OpCode      : 0x88 - 0x8Fh, 0x98 - 0x9Fh    */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(d9,b3)=0      */
/*             : then (PC)<-(PC) + r8          */
/* Flags Used  : None                          */
/***********************************************/
int BN(unsigned char Opcode)
{
  unsigned short OldPC = Cpu.PC + 1;
  int Bit = Opcode & 0x7;
  unsigned char Data = ReadByte(((Opcode & 0x10) << 4) | Cpu.Code[OldPC]);

  if (Data & (1 << Bit))
  {
    Cpu.PC += 3;
  }
  else
  {
    Cpu.PC += 3;
    Cpu.PC += (char)Cpu.Code[OldPC + 1];
  }

  return 2;
}

/* DBNZ d9,r8 (Decrement direct byte and branch near relative address if direct byte is not zero) */
/***********************************************/
/* OpCode      : 0x52 - 0x53h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,(d9)=(d9)-1      */
/*             : if(d9)!=0then(PC)<-(PC)+r8    */
/* Flags Used  : None                          */
/***********************************************/
int DBNZd9(unsigned char Opcode)
{
  /* NOTE: Port latch !!! */

  unsigned short Addr = ((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1];
  unsigned char  Data = ReadByte(Addr) - 1;
  WriteByte(Addr, Data);

  if (Data != 0)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
  }
  else
  {
    Cpu.PC += 3;
  }

  return 2;
}

/* DBNZ @Rj,r8 (Decrement indirect byte and branch near relative address if indirect byte is not zero) */
/***********************************************/
/* OpCode      : 0x54 - 0x57h                  */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+2,((Rj))=((Rj))-1  */
/*             : if((Rj))!=0then(PC)<-(PC)+r8  */
/* Flags Used  : None                          */
/***********************************************/
int DBNZrj(unsigned char Opcode)
{
  /* NOTE: Port latch !!! */

  unsigned short Addr = GetIndirectAddr(Opcode & 0x3);
  unsigned char  Data = ReadByte(Addr) - 1;
  WriteByte(Addr, Data);

  if (Data != 0)
  {
    Cpu.PC = Cpu.PC + 2 + (char)Cpu.Code[Cpu.PC + 1];
  }
  else
  {
    Cpu.PC += 2;
  }

  return 2;
}

/* BE #i8,r8 (Compare immediate data to accumulator and branch near relative address if equal) */
/***********************************************/
/* OpCode      : 0x31h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(ACC) == #i8   */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BEi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1];
  
  if (Cpu.Acc == Data)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
    SetCY(0);
  }
  else
  {
    Cpu.PC += 3;
    if (Cpu.Acc < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }

  return 2;
}

/* BE d9,r8 (Compare direct byte to accumulator and branch near relative address if equal) */
/***********************************************/
/* OpCode      : 0x32 - 0x33h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(ACC) == (d9)  */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BEd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);
  
  if (Cpu.Acc == Data)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
    SetCY(0);
  }
  else
  {
    Cpu.PC += 3;
    if (Cpu.Acc < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }

  return 2;
}

/* BE @Rj,#i8,r8 (Compare immediate data to indirect byte and branch near relative address if equal) */
/***********************************************/
/* OpCode      : 0x34 - 0x37h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if((Rj)) == #i8  */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BErj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3));
  unsigned char Byte = Cpu.Code[Cpu.PC + 1];
  
  if (Byte == Data)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
    SetCY(0);
  }
  else
  {
    Cpu.PC += 3;
    if (Byte < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }

  return 2;
}

/* BNE #i8,r8 (Compare immediate data to accumulator and branch near relative address if not equal) */
/***********************************************/
/* OpCode      : 0x41h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(ACC) != #i8   */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BNEi8(unsigned char Opcode)
{
  unsigned char Data = Cpu.Code[Cpu.PC + 1];
  
  if (Cpu.Acc != Data) /* Not equal */
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
    
    if (Cpu.Acc < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }
  else /* Equal */
  {
    Cpu.PC += 3;
    SetCY(0);
  }

  return 2;
}

/* BNE d9,r8 (Compare direct byte to accumulator and branch near relative address if not equal) */
/***********************************************/
/* OpCode      : 0x42 - 0x43h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if(ACC) != (d9)  */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BNEd9(unsigned char Opcode)
{
  unsigned char Data = ReadByte(((Opcode & 1) << 8) | Cpu.Code[Cpu.PC + 1]);
  
  if (Cpu.Acc != Data)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];

    if (Cpu.Acc < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }
  else
  {
    Cpu.PC += 3;
    SetCY(0);
  }

  return 2;
}

/* BNE @Rj,#i8,r8 (Compare immediate data to indirect byte and branch near relative address if not equal) */
/***********************************************/
/* OpCode      : 0x44 - 0x47h                  */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : (PC)<-(PC)+3,if((Rj)) != #i8  */
/*             : then (PC)<-(PC)+r8            */
/* Flags Used  : CY                            */
/***********************************************/
int BNErj(unsigned char Opcode)
{
  unsigned char Data = ReadByte(GetIndirectAddr(Opcode & 0x3));
  unsigned char Byte = Cpu.Code[Cpu.PC + 1];
  
  if (Byte != Data)
  {
    Cpu.PC = Cpu.PC + 3 + (char)Cpu.Code[Cpu.PC + 2];
    
    if (Byte < Data) 
    {
      SetCY(1);
    }
    else 
    {
      SetCY(0);
    }
  }
  else
  {
    Cpu.PC += 3;
    SetCY(0);
  }

  return 2;
}

/*************************************/
/****** Subroutine Instructions ******/
/*************************************/

/* CALL a12 (Near absolute subroutine call) */
/***********************************************/
/* OpCode      : 0x08 - 0x0Fh, 0x18 - 0x1Fh    */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 2                             */
/* Function    : See VMU documentation         */
/* Flags Used  : None                          */
/***********************************************/
int CALL(unsigned char Opcode)
{
  unsigned short Data = (((Opcode & 0x10) >> 1) | (Opcode & 0x7)) << 8;
  Data |= Cpu.Code[Cpu.PC + 1];

  Cpu.PC += 2;

  Ram[0][++Cpu.SP] = Cpu.PC & 0xFF;
  Ram[0][++Cpu.SP] = Cpu.PC >> 8;

  Cpu.PC = (Cpu.PC & 0xF000) | (Data & 0x0FFF);

  return 2;
}

/* CALLF a16 (Far absolute subroutine call) */
/***********************************************/
/* OpCode      : 0x20h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 2                             */
/* Function    : See VMU documentation         */
/* Flags Used  : None                          */
/***********************************************/
int CALLF(unsigned char Opcode)
{
  unsigned short Data = (Cpu.Code[Cpu.PC + 1] << 8) | Cpu.Code[Cpu.PC + 2];

  Cpu.PC += 3;

  Ram[0][++Cpu.SP] = Cpu.PC & 0xFF;
  Ram[0][++Cpu.SP] = Cpu.PC >> 8;

  Cpu.PC = Data;

  return 2;
}

/* CALLR r16 (Far relative subroutine call) */
/***********************************************/
/* OpCode      : 0x10h                         */
/* Nr of Bytes : 3                             */
/* Nr of Cycles: 4                             */
/* Function    : See VMU documentation         */
/* Flags Used  : None                          */
/***********************************************/
int CALLR(unsigned char Opcode)
{
  unsigned short Data = (Cpu.Code[Cpu.PC + 2] << 8) | Cpu.Code[Cpu.PC + 1];

  Cpu.PC += 3;

  Ram[0][++Cpu.SP] = Cpu.PC & 0xFF;
  Ram[0][++Cpu.SP] = Cpu.PC >> 8;

  Cpu.PC = (Cpu.PC - 1) + Data;

  return 4;
}

/* RET (Return from subroutine) */
/***********************************************/
/* OpCode      : 0xA0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 2                             */
/* Function    : See VMU documentation         */
/* Flags Used  : None                          */
/***********************************************/
int RET(unsigned char Opcode)
{
  Cpu.PC = (Ram[0][Cpu.SP--] << 8);
  Cpu.PC |= Ram[0][Cpu.SP--];

  return 2;
}

/* RETI (Return from interrupt) */
/***********************************************/
/* OpCode      : 0xB0h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 2                             */
/* Function    : See VMU documentation         */
/* Flags Used  : None                          */
/***********************************************/
int RETI(unsigned char Opcode)
{
  Cpu.PC = (Ram[0][Cpu.SP--] << 8);
  Cpu.PC |= Ram[0][Cpu.SP--];

  IntMask--;

  IntMask &= ~0x80;
    
  POTATO_ProcessInterrupt();

  return 2;
}

/*******************************************/
/****** Bit Manipulation Instructions ******/
/*******************************************/

/* CLR1 d9,b3 (Clear direct bit) */
/***********************************************/
/* OpCode      : 0xC8 - 0xCFh, 0xD8 - 0xDFh    */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9,b3) <- 0                  */
/* Flags Used  : None                          */
/***********************************************/
int CLR1(unsigned char Opcode)
{
  unsigned short Addr = ((Opcode & 0x10) << 4) | Cpu.Code[Cpu.PC + 1];
  unsigned char Data = ReadByte(Addr);
  
  WriteByte(Addr, Data & ~(1 << (Opcode & 0x7)));

  Cpu.PC += 2;
  return 1;
}

/* SET1 d9,b3 (Clear direct bit) */
/***********************************************/
/* OpCode      : 0xE8 - 0xEFh, 0xF8 - 0xFFh    */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9,b3) <- 1                  */
/* Flags Used  : None                          */
/***********************************************/
int SET1(unsigned char Opcode)
{
  unsigned short Addr = ((Opcode & 0x10) << 4) | Cpu.Code[Cpu.PC + 1];
  unsigned char Data = ReadByte(Addr);
  
  WriteByte(Addr, Data | (1 << (Opcode & 0x7)));

  Cpu.PC += 2;
  return 1;
}

/* NOT1 d9,b3 (Not direct bit) */
/***********************************************/
/* OpCode      : 0xA8 - 0xAFh, 0xB8 - 0xBFh    */
/* Nr of Bytes : 2                             */
/* Nr of Cycles: 1                             */
/* Function    : (d9,b3) <- (d9,b3)            */
/* Flags Used  : None                          */
/***********************************************/
int NOT1(unsigned char Opcode)
{
  unsigned short Addr = ((Opcode & 0x10) << 4) | Cpu.Code[Cpu.PC + 1];
  unsigned char Data = ReadByte(Addr);
  
  WriteByte(Addr, Data ^ (1 << (Opcode & 0x7)));

  Cpu.PC += 2;
  return 1;
}

/*******************************************/
/****** Miscellaneous Instructions ******/
/*******************************************/

/* NOP (No Operation) */
/***********************************************/
/* OpCode      : 0x00h                         */
/* Nr of Bytes : 1                             */
/* Nr of Cycles: 1                             */
/* Function    :                               */
/* Flags Used  : None                          */
/***********************************************/
int NOP(unsigned char Opcode)
{
  Cpu.PC += 1;
  return 1;
}

/* ILLEGAL */
/***********************************************/
/* OpCode      :                               */
/* Nr of Bytes :                               */
/* Nr of Cycles:                               */
/* Function    :                               */
/* Flags Used  :                               */
/***********************************************/
int ILLEGAL(unsigned char Opcode)
{
  ShowMessage("Illegal Opcode at 0x%x :: 0x%x", Cpu.PC, Opcode);
  
  Cpu.PC += 1;
  return 1;
}