#ifndef __VMU_CORE_H
#define __VMU_CORE_H

#ifdef __cpluplus
extern "C" {
#endif

#define SetCY(n) (Cpu.CY = (n))
#define SetOV(n) (Cpu.OV = (n))
#define SetAC(n) (Cpu.AC = (n))
#define GetPSW() (Cpu.PSW = Cpu.CY << 7 | Cpu.AC << 6 | Cpu.IRBK1 << 4 | Cpu.IRBK0 << 3 | Cpu.OV << 2 | Cpu.RAMBK0 << 1 | Parity[Cpu.Acc]);
#define SetPSW() (Cpu.CY     = ((Cpu.PSW & 0x80) >> 7)); \
                 (Cpu.AC     = ((Cpu.PSW & 0x40) >> 6)); \
                 (Cpu.IRBK1  = ((Cpu.PSW & 0x10) >> 4)); \
                 (Cpu.IRBK0  = ((Cpu.PSW & 0x08) >> 3)); \
                 (Cpu.OV     = ((Cpu.PSW & 0x04) >> 2)); \
                 (Cpu.RAMBK0 = ((Cpu.PSW & 0x02) >> 1)); \
                 (Cpu.P      = ((Parity[Cpu.Acc])    ));  

typedef int VMU_Opcode(unsigned char Opcode);
extern  VMU_Opcode *VMU_OpTable[0x100];

struct _POTATO
{
  unsigned short PC;  /* Program Counter     */
  unsigned char  PSW; /* Program Status Word */
  unsigned char  SP;  /* Stack Pointer       */
  
  unsigned char Acc;  /* Accumulator */
  unsigned char B;    /* B Register  */
  unsigned char C;    /* C Register  */

  int CY;     /* Carry Flag                            */
  int AC;     /* Auxiliary Carry Flag                  */
  int IRBK1;  /* Indirect Address Register Bank Flag 1 */
  int IRBK0;  /* Indirect Address Register Bank Flag 0 */
  int OV;     /* Overflow Flag                         */
  int RAMBK0; /* Data Memory (RAM) Bank Flag           */
  int P;      /* Accumulator Parity                    */

  /* Table Reference Register */
  union
	{
		struct
		{
			unsigned char Low;
			unsigned char High;
		} Byte;
		unsigned short Value;
	} TRR;
  unsigned char FPR; /* FlashRom Programming Register */

  /* Timer 0 */
  struct
  {
    unsigned char T0CNT;  /* Control Register       */
    unsigned char T0PRR;  /* Prescale Data Register */
    
    union
    {
      struct
      {
        unsigned char T0L;  /* Low Register (8-bit)  */
        unsigned char T0H;  /* High Register (8-bit) */
      } Byte;
      unsigned short T0;    /* Register (16-bit)     */
    } Timer;

    union
    {
      struct
      {
        unsigned char T0LR; /* Low Reload Register      */
        unsigned char T0HR; /* High Reload Register     */
      } Byte;
      unsigned short T0R;   /* Reload Register (16-bit) */
    } Reload;

    int TimeBase;
    int PreScale;

  } Timer0;

  struct
  {
    unsigned char  BTCR;    /* Base Timer Control Register */
    unsigned short Counter; /* 14-Bits counter             */
  } BaseTimer;

  unsigned char PCON; /* Power Control Regitser */
  unsigned char IE;   /* Master Interrupt Enable Control Register */
  unsigned char IP;   /* Interrupt Priority Control Register */
  unsigned char EXT;  /* External Memory Control Register */
  unsigned char OCR;  /* Oscillation Control Register */
  
  unsigned char MCR;  /* Mode Control Register         */
  unsigned char XBNK; /* Bank Address Register         */
  unsigned char VCCR; /* LCD Contrast Control Register */

  unsigned char P1;     /* Port 1 Latch */
  unsigned char P3;     /* Port 3 Latch */
  unsigned char P3INT;  /* Port 3 Interrupt Control Register */
  unsigned char P7;     /* Port 7 Latch */

  unsigned char VSEL;   /* VMU Control Register */
  unsigned char VRMAD1; /* Work Ram Access Address 1 */
  unsigned char VRMAD2; /* Work Ram Access Address 2 */

  unsigned char* Code;    /* Instruction Code Pointer  */
  int            Speed;   /* Current Clock Speed */
  int            NrCycles;
  int            Halt;
  int            Hold;
};

extern _POTATO Cpu;
extern unsigned char BiosRom[0x10000];
extern unsigned char FlashRom[2][0x10000];
extern unsigned char Ram[2][0x100];
extern unsigned char WorkRam[2][0x100];
extern unsigned char XRam[3][0x80];
extern unsigned char Parity[0x100];

extern int IntMask;
extern int IntReq;

/* CPU Functions */
void POTATO_Init();

void POTATO_HardReset();
void POTATO_SoftReset();

void POTATO_Execute(int Instructions);
void POTATO_ExecuteCycles(int Cycles);

void POTATO_RunTimers(int Cycles);

void POTATO_Interrupt(int Level);
void POTATO_ProcessInterrupt();

int  POTATO_LoadVms(char *Name);
int  POTATO_LoadBios(char *Name);
int  POTATO_Disassemble(unsigned short PC, char *TextBuffer);

/* Data Map Functions */
unsigned char ReadByte(unsigned short Addr);
void WriteByte(unsigned short Addr, unsigned char Data);

/* LCD Functions */
void POTATO_DrawLCD();

/* Controller (P3) Functions */
void POTATO_PressKey(int Key);
void POTATO_ReleaseKey(int Key);

/* Opcodes */
int ADDi8(unsigned char Opcode);
int ADDd9(unsigned char Opcode);
int ADDrj(unsigned char Opcode);
int ADDCi8(unsigned char Opcode);
int ADDCd9(unsigned char Opcode);
int ADDCrj(unsigned char Opcode);
int SUBi8(unsigned char Opcode);
int SUBd9(unsigned char Opcode);
int SUBrj(unsigned char Opcode);
int SUBCi8(unsigned char Opcode);
int SUBCd9(unsigned char Opcode);
int SUBCrj(unsigned char Opcode);
int INCd9(unsigned char Opcode);
int INCrj(unsigned char Opcode);
int DECd9(unsigned char Opcode);
int DECrj(unsigned char Opcode);
int MUL(unsigned char Opcode);
int DIV(unsigned char Opcode);
int ANDi8(unsigned char Opcode);
int ANDd9(unsigned char Opcode);
int ANDrj(unsigned char Opcode);
int ORi8(unsigned char Opcode);
int ORd9(unsigned char Opcode);
int ORrj(unsigned char Opcode);
int XORi8(unsigned char Opcode);
int XORd9(unsigned char Opcode);
int XORrj(unsigned char Opcode);
int ROL(unsigned char Opcode);
int ROLC(unsigned char Opcode);
int ROR(unsigned char Opcode);
int RORC(unsigned char Opcode);
int LDd9(unsigned char Opcode);
int LDrj(unsigned char Opcode);
int STd9(unsigned char Opcode);
int STrj(unsigned char Opcode);
int MOVi8d9(unsigned char Opcode);
int MOVi8rj(unsigned char Opcode);
int LDC(unsigned char Opcode);
int LDF(unsigned char Opcode);
int STF(unsigned char Opcode);
int PUSH(unsigned char Opcode);
int POP(unsigned char Opcode);
int XCHd9(unsigned char Opcode);
int XCHrj(unsigned char Opcode);
int JMP(unsigned char Opcode);
int JMPF(unsigned char Opcode);
int BR(unsigned char Opcode);
int BRF(unsigned char Opcode);
int BZ(unsigned char Opcode);
int BNZ(unsigned char Opcode);
int BP(unsigned char Opcode);
int BPC(unsigned char Opcode);
int BN(unsigned char Opcode);
int DBNZd9(unsigned char Opcode);
int DBNZrj(unsigned char Opcode);
int BEi8(unsigned char Opcode);
int BEd9(unsigned char Opcode);
int BErj(unsigned char Opcode);
int BNEi8(unsigned char Opcode);
int BNEd9(unsigned char Opcode);
int BNErj(unsigned char Opcode);
int CALL(unsigned char Opcode);
int CALLF(unsigned char Opcode);
int CALLR(unsigned char Opcode);
int RET(unsigned char Opcode);
int RETI(unsigned char Opcode);
int CLR1(unsigned char Opcode);
int SET1(unsigned char Opcode);
int NOT1(unsigned char Opcode);
int NOP(unsigned char Opcode);
int ILLEGAL(unsigned char Opcode);

#ifdef __cpluplus
extern };
#endif

#endif /* __VMU_CORE_H */